unit PMX.Importer;

interface

uses
  System.SysUtils, System.Classes, System.UITypes, System.UIConsts, FMX.Types,
  FMX.Types3D, FMX.Import, FMX.Objects3D, FMX.MaterialSources, FMX.Forms,
  System.Generics.Collections{$IF FireMonkeyVersion >= 19}, FMX.Graphics{$IFEND};

type
  TPMXModelImporter = class(TModelImporter)
  public
    function GetDescription: string; override;
    function GetExt: string; override;
    function LoadFromFile(const AFileName: string; out AMesh: TMeshDynArray;
      const AOwner: TComponent): Boolean; override;
  end;

implementation

uses
  PMX.Model;

{ TPMXModelImporter }

function TPMXModelImporter.GetDescription: string;
begin
  Result := 'MikuMikuDance Model';
end;

function TPMXModelImporter.GetExt: string;
begin
  Result := 'PMX';
end;

function TPMXModelImporter.LoadFromFile(const AFileName: string;
  out AMesh: TMeshDynArray; const AOwner: TComponent): Boolean;
var
  pmx: TPmx;
  path: string;
  iMat, iVert, nFace, iOldIdx, iNewIdx: Integer;
  mesh: TMesh;
  vertices, vertices_part: array of TMeshVertex;
  indices: array of Cardinal;
  material: TLightMaterialSource;
  root: TComponent;
  dictIndices: TDictionary<Integer,Integer>;

  function GetRelocatedIndex(Index: Integer): Integer;
  begin
    if not dictIndices.TryGetValue(Index, Result) then
    begin
      Result := dictIndices.Count;
      dictIndices.Add(Index, Result);
    end;
  end;

begin
  if not FileExists(AFileName) then Exit(False);
  path := ExtractFilePath(AFileName);

  pmx := TPmx.Create;
  try
    pmx.LoadFromFile(AFileName, False, 1.0);

    SetLength(vertices, pmx.VertexCount);
    for iVert := 0 to pmx.VertexCount-1 do
    begin
      vertices[iVert].x := pmx.Vertices[iVert].Position.X;
      vertices[iVert].y := -pmx.Vertices[iVert].Position.Y;
      vertices[iVert].z := pmx.Vertices[iVert].Position.Z;
      vertices[iVert].nx := pmx.Vertices[iVert].Normal.X;
      vertices[iVert].ny := -pmx.Vertices[iVert].Normal.Y;
      vertices[iVert].nz := pmx.Vertices[iVert].Normal.Z;
      vertices[iVert].tu := pmx.Vertices[iVert].UV.X;
      vertices[iVert].tv := pmx.Vertices[iVert].UV.Y;
    end;

    dictIndices := TDictionary<Integer,Integer>.Create;
    try
      nFace := 0;
      SetLength(AMesh, pmx.MaterialCount);
      for iMat := 0 to pmx.MaterialCount-1 do
      begin
        dictIndices.Clear;
        SetLength(indices, pmx.Materials[iMat].NumberOfVertex);
        for iVert := 0 to pmx.Materials[iMat].NumberOfVertex div 3-1 do
        begin
          indices[iVert*3+0] := GetRelocatedIndex(pmx.Faces[nFace + iVert]^.X);
          indices[iVert*3+1] := GetRelocatedIndex(pmx.Faces[nFace + iVert]^.Y);
          indices[iVert*3+2] := GetRelocatedIndex(pmx.Faces[nFace + iVert]^.Z);
        end;
        SetLength(vertices_part, dictIndices.Count);
        for iOldIdx in dictIndices.Keys do
        begin
          iNewIdx := dictIndices[iOldIdx];
          vertices_part[iNewIdx] := vertices[iOldIdx];
        end;

        mesh := TMesh.Create(AOwner);

        if Assigned(mesh.Root) then
          root := mesh.Root.GetObject
        else if Assigned(AOwner.Owner) then
          root := AOwner.Owner
        else
          root := AOwner;
        material := TLightMaterialSource.Create(root);
        {$IFDEF MSWINDOWS}
        if (csDesigning in AOwner.ComponentState) and (root is TCommonCustomForm) then
          material.Name := TCommonCustomForm(root).Designer.UniqueName(AOwner.Name + '_Mat' + Format('%.3d', [iMat]));
        {$ENDIF}
        with pmx.Materials[iMat].Diffuse do
          material.Diffuse := MakeColor(Trunc(X * 255), Trunc(Y * 255), Trunc(Z * 255), Trunc(W * 255));
        with pmx.Materials[iMat].Ambient do
          material.Ambient := MakeColor(Trunc(X * 255), Trunc(Y * 255), Trunc(Z * 255));
        with pmx.Materials[iMat].Specular do
          material.Specular := MakeColor(Trunc(X * 255), Trunc(Y * 255), Trunc(Z * 255));
        material.Shininess := Trunc(pmx.Materials[iMat].Shininess);

        if (pmx.Materials[iMat].TextureIndex <> -1) and FileExists(path + pmx.Textures[pmx.Materials[iMat].TextureIndex])  then
        begin
          material.Texture.LoadFromFile(path + pmx.Textures[pmx.Materials[iMat].TextureIndex]);
        end;

        mesh.MaterialSource := material;

        //mesh.Data.ChangeFormat([TVertexFormat.vfVertex, TVertexFormat.vfNormal, TVertexFormat.vfTexCoord0]);
        mesh.Data.AssignFromMeshVertex(vertices_part, indices);
        mesh.Locked := True;
        AMesh[iMat] := mesh;

        Inc(nFace, pmx.Materials[iMat].NumberOfVertex div 3);
      end;
    finally
      dictIndices.Free;
    end;

    Result := True;
  finally
    pmx.Free;
  end;
end;

var
  PMXImporterId: Integer;
initialization
  PMXImporterId := TModelImportServices.RegisterImporter(TPMXModelImporter.Create);
finalization
  TModelImportServices.UnregisterImporter(PMXImporterId);
end.
